#!/usr/bin/perl -w

# Gianduia, DGAS job monitor and Usage Records collector.
# written by R.M. Piro (piro@to.infn.it) with code
# taken from A. Guarise (guarise@to.infn.it)
# and G. Patania (patania@to.infn.it).

use strict;
use POSIX;
use Text::ParseWords;
use Time::Local;
use File::Basename;


# turn off buffering of STDOUT
$| = 1;


my $TSTAMP_ACC = 86400; # timestamps between CE log and LRMS log can
                        # differ for up to a day.
my $DEF_LDIF_VALIDITY = 86400; # assume the GLUE attributes taken from the LDIF
                               # didn't change within the last day.

my $DEF_IGNORE_JOBS_LOGGED_BEFORE = "2006-01-01";

my $sigset = POSIX::SigSet ->new();
my $actionHUP = 
    POSIX::SigAction->new("sigHUP_handler",$sigset,&POSIX::SA_NODEFER);
my $actionInt = 
    POSIX::SigAction->new("sigINT_handler",$sigset,&POSIX::SA_NODEFER);
POSIX::sigaction(&POSIX::SIGHUP, $actionHUP);
POSIX::sigaction(&POSIX::SIGINT, $actionInt);
POSIX::sigaction(&POSIX::SIGTERM, $actionInt);
my $configFilePath ="/opt/glite/etc/dgas_gianduia.conf";
my %configValues =
    (
     lrmsType              =>  "",
     pbsAcctLogDir         =>  "",
     lsfAcctLogDir         =>  "",
     ceJobMapLog          =>  "",
     useCEJobMap          =>  "yes",
     jobsToProcess        =>  "all",

     keyList               => "GlueHostBenchmarkSF00,GlueHostBenchmarkSI00",
     ldifDefaultFiles      => "",
     glueLdifFile          => "",

     gianduiottiBox    =>  "/opt/glite/var/dgasURbox",
     
     collectorLockFileName => "/opt/glite/var/dgas_gianduia_urCollector.lock",
     collectorBufferFileName => "/opt/glite/var/dgasCollectorBuffer",

     mainPollInterval  => "5",
     timeInterval      =>"5",
     jobPerTimeInterval =>"10",

     ignoreJobsLoggedBefore => $DEF_IGNORE_JOBS_LOGGED_BEFORE,
     );


my $onlyOneIteration = 0; # default is run as daemon!
my $useCElog = 1;   # default is use the CE's map: grid job <-> local job
my $processGridJobs = 1; # default, changed through jobsToProcess
my $processLocalJobs = 1; # default, changed through jobsToProcess

# get command line arguments (if any):
my $clinearg = "";
while (@ARGV) {
    $clinearg = shift @ARGV;
    if ($clinearg eq "--nodaemon") {
	$onlyOneIteration = 1; # one iteration then quit! (for cronjobs)
    }
    else {
	# take it as configuration file name
	$configFilePath = $clinearg;
    }
}



# Parse configuration file
&parseConf($configFilePath);

my $lrmsType = $configValues{lrmsType};
my $pbsAcctLogDir = $configValues{pbsAcctLogDir};
my $lsfAcctLogDir = $configValues{lsfAcctLogDir};
my $ceJobMapLog = $configValues{ceJobMapLog};
my $useCEJobMap = $configValues{useCEJobMap};
my $jobsToProcess = $configValues{jobsToProcess};
my $keyList = $configValues{keyList};
my $ldifDefaultFiles = $configValues{ldifDefaultFiles};
my $glueLdifFile = $configValues{glueLdifFile};
my $gianduiottiBox   = $configValues{gianduiottiBox};
my $collectorLockFileName = $configValues{collectorLockFileName};
my $collectorBufferFileName = $configValues{collectorBufferFileName};
my $mainPollInterval = $configValues{mainPollInterval};
my $timeInterval = $configValues{timeInterval};
my $jobPerTimeInterval = $configValues{jobPerTimeInterval};


# default: ignore jobs before 2006-01-01 00:00:00 UTC
# check that UR box exists (where the pushd expects the URs), if not create it!
( -d $gianduiottiBox ) || mkdir $gianduiottiBox;
chmod 0750, $gianduiottiBox;



# put lock
if ( &putLock($collectorLockFileName) != 0 ) {
    &errorNoRemoveLock("Fatal Error: Couldn't open lock file! in $collectorLockFileName\nExiting ...\n");
}
else {
    print "".localtime().": Daemon started. Lock file succesfully created.\n";
}


if ($jobsToProcess =~ /^grid$/i) {
    $processGridJobs = 1; 
    $processLocalJobs = 0; # only grid jobs
    print "Processing only _grid_ jobs (with entries in ceJobMapLog)!\n";
}
elsif ($jobsToProcess =~ /^local$/i) {
    $processGridJobs = 0; # only local jobs
    $processLocalJobs = 1;
    print "Processing only _local_ jobs (without entries in ceJobMapLog)!\n";
}
elsif ($jobsToProcess =~ /^all$/i) {
    $processGridJobs = 1; # default
    $processLocalJobs = 1; # default
    print "Processing _all_ jobs (grid and local)!\n";
}
else {
    print "Warning: In configuration file: cannot regonize 'jobsToProcess = \"$jobsToProcess\"', possible values are: \"grid\", \"local\" and \"all\". Using default: \"all\"!\n";
    $processGridJobs = 1; # default
    $processLocalJobs = 1; # default
    print "Processing _all_ jobs (grid and local)!\n";
}



# ignore old jobs? default is: ignore before $DEF_IGNORE_JOBS_LOGGED_BEFORE
if ($configValues{ignoreJobsLoggedBefore} !~ /^(\d{4})[-\/](\d{2})[-\/](\d{2})$/) {
    print "Attention: Cannot interpret value specified for ignoreJobsLoggedBefore in configuration file: \"$configValues{ignoreJobsLoggedBefore}\" ... using default!\n";
    $configValues{ignoreJobsLoggedBefore} = $DEF_IGNORE_JOBS_LOGGED_BEFORE;
}

my $ignoreJobsLoggedBefore = &date2Timestamp($configValues{ignoreJobsLoggedBefore});

print "Ignoring jobs whose usage was logged before $configValues{ignoreJobsLoggedBefore} 0:00 am (in UTC => $ignoreJobsLoggedBefore)!\n";



# check wether to use the CE log and get name of directory:
my $ceJobMapLogDir = "";

if ($useCEJobMap ne "yes" && $useCEJobMap ne "YES") {
    $useCElog = 0;  # don't use it, treat all jobs as local
    print "Warning: Not using the CE's job map log file for retrieving grid-related\ninformation. All jobs treated as local jobs!\n";

    # must be set if processing _only_ grid jobs or _only_ local jobs
    if (!$processGridJobs || !$processLocalJobs) {
	&error("Error in configuration file: 'jobsToProcess = \"$jobsToProcess\"' requires useCEJobMap and ceJobMapLog to be set!\n");
    }

}
else {
    if ( -d $ceJobMapLog ) {
	# specified in conf file: directory, not a file
	$ceJobMapLogDir = $ceJobMapLog;
    }
    else {
	# if $ceJobMapLog is a file, not a directory, get the directory:
	$ceJobMapLogDir = dirname($ceJobMapLog)."/";
    }

    # check whether what was specified exists:
    if ( ! -d $ceJobMapLogDir ) {
	&error("Fatal error: directory with ceJobMapLog doesn't exist: '$ceJobMapLog'\n");
    }
}


my $timeZone = `date +%z`;
chomp($timeZone);
my $domainName = `hostname -d`;
chomp($domainName);

# This is for parsing LDIF files:
my %glueAttributes = ();
my $ldifModTimestamp = 0;

my $keepgoing = 1;
my $jobsThisStep = 0; # used to process only bunchs of jobs, as specified in
                      # the configuration file

# determine the LRMS type we have to treat:
my $lrmsLogDir = "";
if ($lrmsType eq "pbs") {
    $lrmsLogDir = $pbsAcctLogDir;
}
elsif ($lrmsType eq "lsf") {
    $lrmsLogDir = $lsfAcctLogDir;
}
elsif ($lrmsType eq "") {
    &error("Error: LRMS type not specified in configuration file!\n");
}
else {
    &error("Error: Unknown LRMS type specified in configuration file!\n");
}


# check LRMS log directory:
if (! -d $lrmsLogDir || ! -r $lrmsLogDir || ! -x $lrmsLogDir) {
    &error("Error: Directory for LRMS log ($lrmsLogDir) cannot be accessed!\n");
}


# append LRMS type to buffer file name:
$collectorBufferFileName = $collectorBufferFileName.".".$lrmsType;


# check whether we can use the 'less' command (should be present on each
# Linux system ...) this is usefull for gzipped log files!
my $have_less = 1;
my $less_cmd = `which less`;
chomp($less_cmd);
if ( ! -x $less_cmd ) {
    $have_less = 0;
}

# this is used for parsing log files backwards ...
my $have_tac = 1;
my $tac_cmd = `which tac`;
chomp($tac_cmd);
if ( ! -x $tac_cmd ) {
    $have_tac = 0;
}

# this is used for parsing log file forwards ...
my $have_cat = 1;
my $cat_cmd = `which cat`;
chomp($cat_cmd);
if ( ! -x $cat_cmd ) {
    $have_cat = 0;
}

if (!$have_tac || !$have_cat) {
    &error("Error: commands 'cat' and 'tac' required for parsing log files! Quitting!\n");
}

if (!$have_less) {
    print "Warning: command 'less' required for parsing compressed log files! Compressed log files will be skipped!\n";
}


# first get info on last job processed:
my $startJob = "";
my $startTimestamp = 0;
my $lastJob = "";
my $lastTimestamp = 0;
my $tmpBuffTstamp = 0;
if (&readBuffer($collectorBufferFileName, $startJob, $startTimestamp) != 0) {
    # we have no regular buffer containing the start Job
    # look for a temporary containing the last jobt + temporary timestamp!
    if (&readTmpBuffer($collectorBufferFileName, $lastJob, $lastTimestamp,
		       $tmpBuffTstamp) != 0) {
	# this is the very first run, not even a temporary buffer is written:
	$tmpBuffTstamp = -1;  # will later be set to first log event timestamp
	                     # encountered!
    }
}

while ($keepgoing) {

    if ($onlyOneIteration) {
	print "".localtime().": Not run as a daemon. Executing a single iteration!\n";
    }

    # see whether the GLUE attributes are available and have changed
    $ldifModTimestamp = &getGLUEAttributesFromLdif();

    # process LRMS log from last file to startJob
    &processLrmsLogs($startJob, $startTimestamp, $lastJob, $lastTimestamp,
		     $tmpBuffTstamp, $ignoreJobsLoggedBefore);

    # write buffer (lastJob and lastTimestamp), if necessary!
    if ( $keepgoing && ($lastJob ne "") && ($lastJob ne $startJob) ) {
	print "".localtime().": Processed backwards from $lastJob (timestamp $lastTimestamp) to $startJob (timestamp $startTimestamp). Updating buffer $collectorBufferFileName.\n";

	&putBuffer($collectorBufferFileName, $lastJob, $lastTimestamp);
	$tmpBuffTstamp = 0;

	$startJob = $lastJob;
	$startTimestamp = $lastTimestamp;
    }

    if ($onlyOneIteration) {
	# not run as a daemon, stop after first round!
	print "".localtime().": Not run as a daemon. Single iteration completed!\n";
	$keepgoing = 0;
    }

    if ($keepgoing) {
	# print "".localtime().": Waiting for new jobs to finish. Sleeping for $mainPollInterval seconds.\n";
	#sleep $mainPollInterval;
	my $secsWaited = 0;
	while ($keepgoing && $secsWaited < $mainPollInterval) {
	    sleep 1;
	    $secsWaited++;
	}
    }

}

## at the end eventually update the buffer
#if ($lastJob ne "" ) {
#    &putBuffer($collectorBufferFileName, $lastJob,
#	       $lastTimestamp);
#}




print "Exiting...\n";
if ( &delLock($collectorLockFileName) != 0 ) {	
    print "Error removing lock file.\n";
}
else {
    print "Lock file removed.\n";
}
print "".localtime().": Exit.\n";

exit(0);





#### ------------------------ END OF MAIN PART ------------------------ ####


#### ------------------------ Functions: ------------------------ ####


#### parseConf ####
sub parseConf {
    my $fconf = $_[0];
    open(FILE, "$fconf") || &errorNoRemoveLock("Error: Cannot open configuration file $fconf\n");
    while(<FILE>) {
	if(/^lrmsType\s*=\s*\"(.*)\"$/){$configValues{lrmsType}=$1;}
	if(/^pbsAcctLogDir\s*=\s*\"(.*)\"$/){$configValues{pbsAcctLogDir}=$1;}
	if(/^lsfAcctLogDir\s*=\s*\"(.*)\"$/){$configValues{lsfAcctLogDir}=$1;}
	if(/^ceJobMapLog\s*=\s*\"(.*)\"$/){$configValues{ceJobMapLog}=$1;}
	if(/^useCEJobMap\s*=\s*\"(.*)\"$/){$configValues{useCEJobMap}=$1;}
	if(/^jobsToProcess\s*=\s*\"(.*)\"$/){$configValues{jobsToProcess}=$1;}

	if(/^keyList\s*=\s*\"(.*)\"$/){$configValues{keyList}=$1;}
	if(/^ldifDefaultFiles\s*=\s*\"(.*)\"$/){$configValues{ldifDefaultFiles}=$1;}
	if(/^glueLdifFile\s*=\s*\"(.*)\"$/){$configValues{glueLdifFile}=$1;}

	if(/^gianduiottiBox\s*=\s*\"(.*)\"$/){$configValues{gianduiottiBox}=$1;}

	if(/^collectorLockFileName\s*=\s*\"(.*)\"$/){$configValues{collectorLockFileName}=$1;}
	if(/^collectorBufferFileName\s*=\s*\"(.*)\"$/){$configValues{collectorBufferFileName}=$1;}
	if(/^mainPollInterval\s*=\s*\"(.*)\"$/){$configValues{mainPollInterval}=$1;}
	if(/^timeInterval\s*=\s*\"(.*)\"$/){$configValues{timeInterval}=$1;}
	if(/^jobPerTimeInterval\s*=\s*\"(.*)\"$/){$configValues{jobPerTimeInterval}=$1;}
	if(/^ignoreJobsLoggedBefore\s*=\s*\"(.*)\"$/){$configValues{ignoreJobsLoggedBefore}=$1;}
    }

    close(FILE);
}



##-------> lock  subroutines <---------##
sub putLock {
    my $lockName = $_[0];
    open(IN,  "< $lockName") && return 1;
    close(IN);
    open(OUT, "> $lockName") || return 2;
    print OUT  $$;    ## writes pid   
    close(OUT);
    return 0;
}

sub delLock {
    my $lockName = $_[0];
    open(IN,  "< $lockName") || return 1;
    close(IN);
    my $status = system("rm -f $lockName");
    return $status;
}

##--------> routines for job processing buffer <---------##
sub putBuffer {
    # arguments are: 0 = buffer name
    #                1 = last LRMS job id
    #                2 = last LRMS job timestamp (log time)
    my $buffName = $_[0];
    if ($keepgoing == 1) {
	# this is done only if no SIGINT was received!
	#print "".localtime().": Writing info on last processed job in buffer $buffName.\n";
	open(OUT, "> $buffName") || return 2;
	print OUT  "$_[1]:$_[2]";
	close(OUT);

	my $tmpBuffer = "${buffName}_tmp";
	if ( -e $tmpBuffer ) {
	    print "Removing temporary buffer $tmpBuffer ... not required anymore!\n";
	    `rm -f $tmpBuffer &> /dev/null`;
	}
    }
    return 0;
}


sub readBuffer {
	my $buffname = $_[0];
	open(IN, "< $buffname") || return 2;
	my $line;
	my $tstamp;
	while ( <IN> )
	{
		($line,$tstamp) = split(':');
		chomp($tstamp); # remove eventual newline
	}	
	close(IN);
	print "Reading buffer $buffname. First job to analyse: id=$line; log timestamp=$tstamp\n";

	$_[1] = $line;
	$_[2] = $tstamp;

	return 0;
}


sub putTmpBuffer {
    # arguments are: 0 = buffer name
    #                1 = last LRMS job id (to go into 
    #                2 = last LRMS job timestamp (log time)
    #                3 = temporary LRMS job timestamp (log time)
    my $buffName = $_[0]."_tmp";
    open(OUT, "> $buffName") || return 2;
    print OUT  "$_[1]:$_[2]:$_[3]";
    close(OUT);

    print "Writing temporary buffer $buffName with '$_[1]:$_[2]:$_[3]'\n";

    return 0;
}


sub readTmpBuffer {
	my $buffname = $_[0]."_tmp";
	open(IN, "< $buffname") || return 2;
	my $tstamp;
	my $startJobId = "";
	my $startJobTstamp = 0;
	my $tmpJobTstamp = 0;
	while ( <IN> )
	{
		($startJobId,$startJobTstamp,$tmpJobTstamp) = split(':');
		chomp($tmpJobTstamp); # remove eventual newline
	}	
	close(IN);
	print "Reading TEMPORARY buffer $buffname (happens if the service was stopped before completing the first run). First job to analyse would be: id=$startJobId; log timestamp=$startJobTstamp. So far processed backwards from there to timestamp: $tmpJobTstamp\n";

	$_[1] = $startJobId;
	$_[2] = $startJobTstamp;
	$_[3] = $tmpJobTstamp;

	return 0;
}


##-------> sig handlers subroutines <---------##
sub sigHUP_handler {
        print "".localtime().": Got SIGHUP!\n";
	$keepgoing = 1;
}

sub sigINT_handler {
        print "".localtime().": Got SIGINT!\n";
        $keepgoing = 0;
}


##--------> process the LRMS log and process the jobs <-------##

# process LRMS log from last file (set lastJob and lastTimestamp) to 
# startJob (get directory in temporal order and check modification date)

sub processLrmsLogs {
    my $startJob = $_[0];
    my $startTimestamp = $_[1];
    my $lastJob = $_[2];
    my $lastTimestamp = $_[3];
    my $tmpBuffTstamp = $_[4];
    my $ignoreJobsLoggedBefore = $_[5];

    my $currLogTimestamp = 0;

    my $continueProcessing = 1;
    my $newestFile = 1; # the first file we open is the current file

    my %processedLogFileInodes = ();

    while ($keepgoing && $continueProcessing) {

	# first get log files and timestamps:
	my @lrmsLogFiles;
	my %logFMod = ();
	my %logFInodes = ();
	my %logFSizes = ();

	my $nothingProcessed = 1; # first assume process nothing in this iteration
	my $allProcessed = 0;

	opendir(DIR, $lrmsLogDir) || &error("Error: can't open dir $lrmsLogDir: $!");
	while( defined(my $file = readdir(DIR)) ) {
	    next if ( $file =~ /^\.\.?$/ ); # skip '.' and '..' 
	    next if ( $lrmsType eq "pbs" && !($file =~ /^\d{8}(\.gz)?$/) );
	    next if ( $lrmsType eq "lsf" && !($file =~ /^lsb\.acct(\.\d*)?(\.gz)?$/) );
	    # we accept compressed files as well (but will be able to parse
	    # them only if we have the command less, see later)

	    push @lrmsLogFiles, $file;

	    # keep track of last modification timestamp:

	    my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev,
		$size,
	        $atime, $mtime, $ctime, $blksize, $blocks); # these are dummies

	    # only inode, size and modification timestamp are interesting!
	    ($dev, $logFInodes{$file}, $mode, $nlink, $uid, $gid, $rdev,
	     $logFSizes{$file},
	     $atime, $logFMod{$file}, $ctime, $blksize, $blocks)
	        = stat("$lrmsLogDir/$file");

	    #$logFMod{$file} = (stat("$lrmsLogDir/$file"))[9]; #mtime in sec.
	    #$logFInodes{$file} = (stat("$lrmsLogDir/$file"))[1]; #inode
	    #$logFSizes{$file} = (stat("$lrmsLogDir/$file"))[7]; #size
	}
	closedir DIR;

	# now we sort the LRMS log files according to their modification
	# timestamp
	my @sortedLrmsLogFiles
	    = (sort{ $logFMod{$b} <=> $logFMod{$a} } keys %logFMod);

	# we process these LRMS log files from the last, until we find the
	# last job previously considered.
	while ($keepgoing && $continueProcessing && @sortedLrmsLogFiles) {

	    my $thisLogFile = shift (@sortedLrmsLogFiles);

	    print "".localtime().": Considering LRMS log file: $thisLogFile; last modified: $logFMod{$thisLogFile}\n";

	    # if this is not the first (newest) log file we read in this
	    # iteration ($currLogTimestamp != 0), make sure we will read only
	    # a log file that contains entries up to the current timestamp! 
	    next if ($currLogTimestamp != 0
		     && $logFMod{$thisLogFile} > $currLogTimestamp);

	    # if a log file with this inode has already been processed in
	    # this iteration ...
	    if (exists($processedLogFileInodes{$logFInodes{$thisLogFile}})) {
		print "File with inode $logFInodes{$thisLogFile} already processed in this iteration, skipping!\n";
		next;
	    }

	    if ($logFSizes{$thisLogFile} == 0) {
		print "File is empty, skipping!\n";
		next;
	    }

	    if ( $logFMod{$thisLogFile} < $startTimestamp
		 || $logFMod{$thisLogFile} < $ignoreJobsLoggedBefore ) {
		# last modified (appended) _before_ the job we have already
		# processed (or: should be ignored because logged to early)
		# hence we stop here!
		$continueProcessing = 0;
	    }
	    else {
		print "".localtime().": Processing LRMS log file: $thisLogFile; last modified: $logFMod{$thisLogFile}\n";

		&processLrmsLogFile($thisLogFile, $newestFile,
				    $_[0], $_[1], $_[2], $_[3], $_[4],
				    $ignoreJobsLoggedBefore, $currLogTimestamp,
				    $nothingProcessed, $allProcessed);

		$processedLogFileInodes{$logFInodes{$thisLogFile}} = 1;

		last; # quit while loop to reload directory for next file!
		      # (file names might have changed due to logrotate!)
	    }
	}

	# if we checked all files and there are none left: stop iteration!
	if (!@sortedLrmsLogFiles) {
	    $continueProcessing = 0;
	}

	# stop iteration also if we either didn't process anything
	# (no new jobs!) or found the last buffer job (processed all new jobs)!
	if ( $allProcessed ||
	    ( $nothingProcessed &&  ($_[0] eq $_[2]) && ($_[1] eq $_[3]) )
	    ) {
	    # $startJob = $_[0]; $startTimestamp = $_[1];
	    # $lastJob = $_[2]; $lastTimestamp = $_[3];
	    $continueProcessing = 0;
	}
    }

}


# --- parse a single log file --- #
sub processLrmsLogFile {

    my $filename = $_[0];
    # my $newestF = $_[1];

    my $startJob = $_[2];
    my $startTimestamp = $_[3];

    my $lastJob = $_[4];
    my $lastTimestamp = $_[5];

    my $tmpBuffTstamp = $_[6];
    my $ignoreJobsLoggedBefore = $_[7];

    my $currLogTimestamp = $_[8];

    my $nothingProcessed = $_[9];
    my $allProcessed = $_[10];

    # for each job to process: check the CE accounting log dir (ordered by
    # modification date) for the right log file +/- 1 (according to last
    # modification date) and process it from backwards.

    # if the job is found in the CE's accounting log: route 2
    # if not: route 3 (local job)

    # building command to open the log file
    my $cmd = $tac_cmd;
    # decide whether to decompress using 'less':
    if ($filename =~ /(\.gz)?$/) {
	# decompress and pipe into tac:
	$cmd = "$less_cmd $lrmsLogDir/$filename | ".$cmd;
    }
    else {
	# just use tac:
	$cmd = $cmd." $lrmsLogDir/$filename";
    }

    #print "".localtime().": Trying to open log file $filename\n";
    if ( !open ( LRMSLOGFILE, "$cmd |" ) ) {
	print "Warning: Couldn't open the log file ... skipping!\n";
	return 1;
    }

    my $firstJobId = 1;

    my $line = "";
    while ($line = <LRMSLOGFILE>) {

	if (!$keepgoing) {
	    print "".localtime().": Stop processing $filename ...\n";
	    close LRMSLOGFILE;
	    return 1;
	}

	if ($line !~ /\n$/) {
	    # no trailing newline: line still imcomplete, currently being
	    # written by the LRMS
	    print "Current line not completed by LRMS, skipping: $line\n";
	    next;
	}

	# not more than a bunch of jobs at a time!
	if ($jobsThisStep == $jobPerTimeInterval) {
	    #print "".localtime().": $jobsThisStep jobs processed. Sleeping for $timeInterval seconds.\n";
	    #sleep $timeInterval;
	    my $secsWaited = 0;
	    while ($keepgoing && $secsWaited < $timeInterval) {
		sleep 1;
		$secsWaited++;
	    }
	    $jobsThisStep = 0;
	}
	# not here, but only when UR file created: $jobsThisStep++;

	# for debugging:
	#print "Scanning line: $line\n";

	# returns an LRMS job ID only if the line contains a finished
	# job
	my $targetJobId = &getLrmsTargetJobId($lrmsType, $line);

	next if ($targetJobId eq "");

	# get event time of LRMS log for the job (0 if not found)
	# this is for the buffer! the CE log timestamp will be matched to 
	# the LRMS creation time (=submission time)!
	my $lrmsEventTimeString = "";
	my $lrmsEventTimestamp =
	    &getLrmsEventTime($lrmsType, $line, $lrmsEventTimeString);

	if($lrmsEventTimestamp == 0) {
	    print "Error: could not determine LRMS event timestamp! Wrong file format ... ignoring this log file!\n";
	    close LRMSLOGFILE;
	    return 1;
	}

	# get creation time stamp for LRMS job (for matching CE log timestamp)
	my $job_ctime = &getLrmsJobCTime($lrmsType, $line);

	if ($job_ctime == 0) {
	    print "Error: could not determine LRMS job creation/submission timestamp! Wrong file format ... ignoring this log file!\n";
	    close LRMSLOGFILE;
	    return 1;
	}


	$_[8] = $currLogTimestamp = $lrmsEventTimestamp;


	if ( $tmpBuffTstamp == 0 && # do this only if we have a regular buffer!
		($targetJobId eq $startJob) && ($startJob ne "") &&
		($lrmsEventTimestamp eq $startTimestamp) && ($startTimestamp ne "")
		) {

	    # write this only if we have processed at least one job!
	    # Otherwise it would be written on all empty iterations.
	    if ( ($_[4] ne $startJob) || ($_[5] ne $startTimestamp)) {
		print "".localtime().": Found already processed $targetJobId with log event time $lrmsEventTimeString (=$lrmsEventTimestamp) in log! Done with iteration!\n";
	    }
	    close LRMSLOGFILE;
	    $_[10] = $allProcessed = 1;
	    return 0;
	}
	elsif ($lrmsEventTimestamp < $ignoreJobsLoggedBefore) {

	    print "".localtime().": Warning: Log event time $lrmsEventTimeString (=$lrmsEventTimestamp) of job $targetJobId BEFORE $ignoreJobsLoggedBefore!\n(This may happen if ignoreJobsLoggedBefore is set in the configuration file and the buffer contained an earlier timestamp)\nStopping iteration!\n";
	    close LRMSLOGFILE;
	    $_[6] = $tmpBuffTstamp = 0;
	    $_[10] = $allProcessed = 1;
	    return 0;
	}
	else {
	    # need to process the job:

	    if ($tmpBuffTstamp < 0) {
		# this is the very first run, we don't even have a temporary 
		# buffer; make sure we start writing temporary buffers until
		# the first iteration is completed:
		print "".localtime().": This is the very first job to process. Start writing a temporary buffer for backward processing until the first iteration is completed!\n";
		$tmpBuffTstamp = $lrmsEventTimestamp;
	    
	    }
	    elsif ($tmpBuffTstamp > 0) {
		# we have a temporary buffer ... process that before creating a
		# new buffer!
		$firstJobId = 0;
	    }


	    # new buffer information should be taken only if we started from
	    # a regular buffer file, not a temporary one!
	    if ( $firstJobId && $_[1] ) {    # $_[1] = $newestF

		$_[4] = $lastJob = $targetJobId; # $lastJob, i.e. newest job processed
		$_[5] = $lastTimestamp = $lrmsEventTimestamp; # $lastTimestamp, i.e. of newest job
		$firstJobId = 0;
		$_[1] = 0; # $newestF; next file (if any) isn't the newest

		print "".localtime().": This is the most recent job to process: $targetJobId; LRMS log event time: $lrmsEventTimeString (=$lrmsEventTimestamp)\n";
	    }

	    if ($tmpBuffTstamp > 0 && $lrmsEventTimestamp > $tmpBuffTstamp) {
		print "".localtime().": Skipping job $targetJobId (since event log timestamp $lrmsEventTimestamp after $tmpBuffTstamp of temporary buffer).\n";
		next;
	    }

	    print "".localtime().": Processing job: $targetJobId with LRMS log event time(local): $lrmsEventTimeString (=$lrmsEventTimestamp); LRMS creation time: $job_ctime\n";


	    $_[9] = $nothingProcessed = 0; # processing something!


	    my $gianduiottoHeader;
	    if ($useCElog) {
		# get grid-related info from CE job map
		$gianduiottoHeader =  &parseCeUserMapLog($targetJobId,
							 $lrmsEventTimestamp,
							 $job_ctime);
	    }
	    else {
		# don't use CE job map ... local job!
		$gianduiottoHeader = "JOB_TYPE=local\n";
	    }

            # keep track of the number of jobs processed:
            $jobsThisStep++;

	    if (!$processLocalJobs && $gianduiottoHeader =~ /JOB_TYPE=local/) {
		print "Skipping this local job (jobsToProcess = \"$jobsToProcess\")!\n";
		next;
	    }
	    elsif (!$processGridJobs && $gianduiottoHeader =~ /JOB_TYPE=grid/) {
		print "Skipping this grid job (jobsToProcess = \"$jobsToProcess\")!\n";
		next;
	    }

	    # eventually add info from LDIF file if this job
	    # is not too old and we can assume the current GLUE
	    # attributes to be correct:
	    if ($ldifModTimestamp != 0) {
		# we add this control later! For now do it always!
		#if ($job_ctime >= $ldifModTimestamp) {
		if (1) {
		    print "Adding GLUE attributes to UR ...\n";
		    my $key;
		    foreach $key (keys(%glueAttributes)) {
			$gianduiottoHeader = $gianduiottoHeader."$key=$glueAttributes{$key}\n";
		    }
		}
		else {
		    print "Job too old, GLUE attributes from LDIF file cannot be assumed to be correct ... not added to UR!\n";
		}
	    }

	    if (&writeGianduiottoFile($targetJobId, $lrmsEventTimestamp,
				      $gianduiottoHeader, $line)
		!= 0) {
		print "".localtime().": Error: could not create UR file in $gianduiottiBox for job $targetJobId with LRMS event time: $lrmsEventTimeString!\n";

		# EXIT BAD!!! don't write an updated buffer!
		close (LRMSLOGFILE);
		&error("Stopping urCollector due to unrecoverable error!\n");
	    }
	    else {
		if ($tmpBuffTstamp != 0) {
		    # in case this is the first run, adjust temporary buffer!
		    $_[6] = $tmpBuffTstamp = $lrmsEventTimestamp;

		    &putTmpBuffer($collectorBufferFileName, $lastJob,
				  $lastTimestamp, $tmpBuffTstamp);
		}
	    }

	    # &parseUR_PBS($_);
	    # print "processing $hash{jobName}\n";
	    # &callAtmClient();
	}

    }  # while (<LRMSLOGFILE>) ...

    close (LRMSLOGFILE);
}



## --------- parse CE user map log file for accounting --------##
## to find a specific job ...

sub parseCeUserMapLog {

    my $lrmsJobID = $_[0];
    my $lrmsTimestamp = $_[1];
    my $job_ctime = $_[2];

    # important note: the CE's log file might not have the precise timestamp
    # of the job's submission to the LRMS (due to implementation problems)
    # hence the CE log's entry is not necessarily in the (rotated?) log file
    # we expect from the LRMS timestamp

    my $gHeader = "";
    my $isLocal = 0; # assume it is a grid job with an entry in the CE log

    my @ceLogFiles;

    my %logFMod = ();

    opendir(DIR, $ceJobMapLogDir) || &error("Error: can't open dir $ceJobMapLogDir: $!");
    while( defined(my $file = readdir(DIR)) ) {
	my $fullname = $ceJobMapLogDir.$file;
	#print "CE_LOG_FILE:$fullname ...\n";

	next if ($file =~ /^\.\.?$/); # skip '.' and '..' 
	next if ( !( $fullname =~ /^$ceJobMapLog[\/\-_]?\d{8}(\.gz)?$/ ) &&
		  !( $fullname =~ /^$ceJobMapLog[\/\-_]?\d{4}-\d{2}-\d{2}(\.gz)?$/ ) &&
		  !( $fullname =~ /^$ceJobMapLog(\.\d*)?(\.gz)?$/)
		  ); # skip if not like "<logname>(-)20060309(.gz)" (one per day)
                     # skip if not like "<logname>(-)2006-03-09(.gz)" (one per day)
                     # and not like "<logname>.1(.gz)" (rotated)!
        push @ceLogFiles, $file;

	#print "accepted!";

	# keep track of last modification timestamp:

	#my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	#    $atime,$mtime,$ctime,$blksize,$blocks)
	#    = stat($file);
	$logFMod{$file} = (stat("$ceJobMapLogDir/$file"))[9]; #mtime in sec. after epoch
    }
    closedir DIR;

    # now we sort the CE log files according to their modification
    # timestamp
    my @sortedCeLogFiles
	= (sort{ $logFMod{$b} <=> $logFMod{$a} } keys %logFMod);

    # find up to 3 file names: the log file that should contain the
    # LRMS timestamp; the previous file and the next file (in case the
    # CE log timestamp is not exactly synchronized the CE log entry might
    # end up in a previous or next file):


    my @ceScanLogFiles = ("", "", "");
    my %scanDirection;

    # $ceScanLogFiles[0] is the file expected for this timestamp
    # $ceScanLogFiles[1] is the previous file
    # $ceScanLogFiles[2] is the next file

    my $logFile = "";
    while (@sortedCeLogFiles) {
	$logFile = shift @sortedCeLogFiles;

	if ($logFMod{$logFile} < $job_ctime) {
	    # this is the first file with an earlier timestamp, thus it is the
	    # previous file
	    $ceScanLogFiles[1] = $logFile;
	    $scanDirection{$logFile} = "backward";
	    last;
	}
	else {
	    # as long as we didn't find the previous file, this might be
	    # the expected one:
	    $ceScanLogFiles[2] = $ceScanLogFiles[0]; # next file
	    $scanDirection{$ceScanLogFiles[2]} = "forward";
	    $ceScanLogFiles[0] = $logFile;           # expected file
	    $scanDirection{$ceScanLogFiles[0]} = "backward";
	}
    }

    # shouldn't be done because due to the skew between the CE log timestamp
    # and the LRMS job_ctime there might be only the previous file in some
    # few cases:
    #
    #if ($ceScanLogFiles[0] eq "") {
    #	print "Warning: no CE user map log file found for the LRMS creation time $job_ctime of job $lrmsJobID!\n";
    #	return "";
    #}
    #else {

    my $scanFile = "";
    my $keepSearchingCeLogs = 1;
    foreach $scanFile (@ceScanLogFiles) {

	last if (!$keepSearchingCeLogs);

	next if ($scanFile eq "");

	print "".localtime().": Scanning CE log $scanFile (last modified=$logFMod{$scanFile}); direction: $scanDirection{$scanFile}\n";

	# decide whether to read forward or backward:
	my $cmd = $tac_cmd;  # default for backward!
	if ($scanDirection{$scanFile} eq "forward") {
	    $cmd = $cat_cmd;
	}
	# decide whether to decompress using 'less':
	if ($scanFile =~ /(\.gz)?$/) {
	    # decompress and pipe into cat/tac:
	    $cmd = "$less_cmd $ceJobMapLogDir/$scanFile | ".$cmd;
	}
	else {
	    # just use cat/tac:
	    $cmd = $cmd." $ceJobMapLogDir/$scanFile";
	}

	if ( !open ( CELOGFILE, "$cmd |" ) ) {
	    print "Warning: Couldn't open the log file ... skipping!\n";
	    next;
	}

	while (my $line = <CELOGFILE>) {
	    if ($line =~ /\s*\"lrmsID=$lrmsJobID\"\s*/) {
		# found something, check timestamp (+/- a day):
		# "timestamp=2006-03-08 12:45:01" or
		# "timestamp=2006/03/08 12:45:01"
		my $ceLogTstamp = 0;
		if ($line =~ /\s*\"timestamp=(\d{4})[-\/](\d{2})[-\/](\d{2})\s(\d{2}):(\d{2}):(\d{2})\"\s*/) {

		    # get timestamp for this UTC time!
		    my $ceEntryTimestamp =
			timegm(int($6),int($5),int($4),    # ss:mm:hh
			       int($3),int($2)-1,int($1)); # dd-mm-yy
		    # month should be from 0 to 11 => -1 !

		    print "".localtime().": Found in CE log: lrmsID=$lrmsJobID with timestamp(UTC)=$1-$2-$3 $4:$5:$6 ($ceEntryTimestamp)\n";

		    if ( ($ceEntryTimestamp > $job_ctime-$TSTAMP_ACC)
			 && ($ceEntryTimestamp < $job_ctime+$TSTAMP_ACC)
			 ) {
			# the timestamp from the CE log is within a day
			# from the LRMS creation timestamp, accept it!

			print "Accepting timestamp from CE log!\nParsing entry: ";

			# example: "timestamp=2006-03-08 12:45:01" "userDN=/C=IT/O=INFN/OU=Personal Certificate/L=Padova/CN=Alessio Gianelle/Email=alessio.gianelle@pd.infn.it" "userFQAN=/atlas/Role=NULL/Capability=NULL" "userFQAN=/atlas/production/Role=NULL/Capability=NULL" "ceID=grid012.ct.infn.it:2119/jobmanager-lcglsf-infinite" "jobID=https://scientific.civ.zcu.cz:9000/-QcMu-Pfv4qHlp2dFvaj9w" "lrmsID=3639.grid012.ct.infn.it"

			# we already got the timestamp and the lrmsID
			my $userDN = "";
			my @userFQANs = ();
			my $ceID = "";
			my $jobID = "";

			my @fields = split(/\"/, $line);
			my $fld;
			foreach $fld (@fields) {
			    next if ($fld =~ /^\s*$/); # spaces in between
			    if ($fld =~ /^userDN=(.*)$/) {
				$userDN = $1;
				print "userDN=$userDN; ";
			    }
			    elsif ($fld =~ /^userFQAN=(.*)$/) {
				my $fqan = $1;
				if (! $fqan =~ /^\s*$/) {
				    push (@userFQANs, $fqan);
				}
				print "userFQAN=$fqan; ";
			    }
			    elsif ($fld =~ /^ceID=(.*)$/) {
				$ceID = $1;
				print "ceID=$ceID; ";
			    }
			    elsif ($fld =~ /^jobID=(.*)$/) {
				$jobID = $1;
				print "jobID=$jobID; ";
				if ($jobID eq "none" || $jobID eq "NONE") {
				    $jobID = "";
				}
			    }
			}
			print "\n";

			# check that we have the minimum info:
			if ($ceID eq "") {
			    print "Warning: ceID missing! Considering this as a local job!\n";
			    $isLocal = 1;
			}

			# info on job
			$gHeader = $gHeader."GRID_JOBID=$jobID\n"
			    ."LRMS_TYPE=$lrmsType\n"
			    ."LRMS_JOBID=$lrmsJobID\n"
			    ."LRMS_EVENTTIME=$lrmsTimestamp\n"
			    ."LRMS_SUBMTIME=$job_ctime\n";

			# info on user
			$gHeader = $gHeader."USER_DN=$userDN\n"
			    ."USER_FQAN=";
			my $fqans = "";
			my $fq = "";
			foreach $fq (@userFQANs) {
			    $fqans .= $fq.";";
			}
			if ($fqans ne "") {
			    chop($fqans); # cut last ";"
			}
			$gHeader = $gHeader.$fqans;

			$gHeader = $gHeader."\n"; # terminate field USER_FQAN

			###$gHeader = $gHeader."HLR_LOCATION=HLR_LOCATION\n";

			# info on CE:
			$gHeader = $gHeader."CE_ID=$ceID\n"
			    ."timeZone=$timeZone\n";

                        $keepSearchingCeLogs = 0;
			last;
		    }
		    elsif ($ceEntryTimestamp<$job_ctime-$TSTAMP_ACC) {
			# the timestamp from the CE log is too low, stop
			# trying to find the job!
			print "Timestamp of CE log before LRMS creation time: no job found in CE log: local job!\n";
			$isLocal = 1;
			$keepSearchingCeLogs = 0;
			last;
		    }
		    else {
			print "Timestamp of CE log after LRMS creation time: job with recycled LRMS ID ... ignoring!\n";
		    }

		} # if ($line =~ /\s*\"timestamp= ...

	    } # if ($line =~ /\s*\"lrmsID= ...

	} # while ($line = <CELOGFILE>) ...

        close CELOGFILE;

    } # foreach ...

    if ($gHeader eq "") {
	print "No job found in CE log: local job!\n";
	$isLocal = 1;
    }

    if ($isLocal) {
	$gHeader = "JOB_TYPE=local\n".$gHeader;
    }
    else {
	# grid job
	$gHeader = "JOB_TYPE=grid\n".$gHeader;
    }

    return $gHeader;
}



## ------ write file in gianduiottiBox per pushd ---------- ##
sub writeGianduiottoFile {

    my $filename = $lrmsType."_".$_[0]."_".$_[1]; 
    # unique filename: <lrmsType>_<lrmsJobID>_<lrmsEventTimestamp>
    my $header = $_[2];
    my $acctlog = $_[3];

    open(OUT, "> $gianduiottiBox/$filename") || return 1;

    print OUT "$header";
    print OUT "ACCTLOG:$acctlog";

    close (OUT);

    return 0;
}


## ------ get GLUE attributes From LDIF file --------- ##
sub getGLUEAttributesFromLdif {
    my $modTStamp = 0; # returns 0 if no valid attributes found!
                       # otherwise returns the timestamp of the last
                       # modification of the file!

    # to be stored in %glueAttributes
    %glueAttributes = (); # first empty everything

    my @ldifFiles = split(/,/, $ldifDefaultFiles);
    unshift(@ldifFiles, $glueLdifFile); # first to try!

    my @keys = split(/,/, $keyList);

    if (!@keys) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no keyList in configuration file)!\n";
	return 0; # no keys -> no benchmarks
    }

    while (@ldifFiles) {
	my $file = shift(@ldifFiles);
	print "".localtime().": Trying to get GLUE benchmarks from LDIF file: $file\n";

	if (!open(GLUEFILE, "< $file")) {
	    print "Warning: could not open the LDIF file ... skipping!\n";
	    next;
	}

	my $foundSomething = 0;
	my $line;
	while ($line = <GLUEFILE>) {
	    my $key;
	    foreach $key (@keys) {
		if ( ($line =~ /^$key:\s?(.*)$/ )
		     || ($line =~ /^$key=(.*)$/ ) 
		     || ($line =~ /^${key}_(.*)$/ ) ) {
		    # accept stuff like "GlueHostBenchmarkSI00: 955" and
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU=955"
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU_955"
		    $glueAttributes{$key} = $1;
		    print "found: $key=$1; ";
		    $foundSomething = 1;
		}
	    }
	}

	close(GLUEFILE);

	if ($foundSomething) {
	    print "\n";
	    # get timestamp of file:
	    $modTStamp = (stat($file))[9]; #mtime in sec. after epoch

	    last; # don't check the other files (if any)!
	}

    }

    my $thisTStamp = time();

    if ($modTStamp == 0) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no valid entries in LDIF file(s))!\n";
	return 0;
    }
    else {
	if($modTStamp > $thisTStamp-$DEF_LDIF_VALIDITY) {
	    # accept it for at least a day, even if more recently modified
	    $modTStamp = $thisTStamp-$DEF_LDIF_VALIDITY;
	}
	# we disabled the controll above!
	#print "GLUE attributes will be accepted for all jobs with LRMS creation time after $modTStamp\n";
    }



    return $modTStamp;
}


## ------ determine the LRMS type from the log file ------ ##
#sub determinLrmsType() {
#    my $lrmsLine = $_[0];
#
#    my $lType = "";  # unknown;
#
#    if ($lrmsLine =~ /^\d{2}\/\d{2}\/\d{4}\s\d{2}:\d{2}:\d{2};.;\d+\./) {
#	# 03/10/2006 00:03:36;S;5738.[t2-ce-01.to.infn.it;user=atlas001 ...]
#	$lType = "pbs";
#    }
#    elsif ($lrmsLine =~ /^\"JOB_FINISH\"\s\"[0-9\.]+\"\s\d+\s/) {
#	# "JOB_FINISH" "6.0" 1140193736 [55455 774 40370307 ...]
#	$lType = "lsf";
#    }
#
#    return $lType;
#}



## ------ these are the LRMS type-specific functions ------ ##


# returns "" if the line should be ignored!
sub getLrmsTargetJobId {
    if ($_[0] eq "pbs") {
	return &getLrmsTargetJobId_pbs($_[1]);
    }
    elsif ($_[0] eq "lsf") {
	return &getLrmsTargetJobId_lsf($_[1]);
    }
    return "";
}

sub getLrmsTargetJobId_pbs {
    my $jid = ""; # default: line to ignore!

    my @ARRAY = split(" " , $_[0] );
    if (scalar(@ARRAY) > 1) {
	my @ARRAY2 = split(";" , $ARRAY[1] );
	if (scalar(@ARRAY2) > 2 && $ARRAY2[1] eq "E" ) {
	    $jid = $ARRAY2[2]; # finished job, return LRMS ID!
	}
    }
    return $jid; # line to ignore?
}

sub getLrmsTargetJobId_lsf {
    my $jid = ""; # default: line to ignore!

    my @ARRAY = split(" " , $_[0] );

    if ( (scalar(@ARRAY) > 3) && ($ARRAY[0] eq "\"JOB_FINISH\"") ) {

	# make sure we consider this line only if it is a _single_ line
	my $recordOK = 1;
	if ($_[0] =~ /^\"JOB_FINISH\".*\"JOB_FINISH\"/) {
	    # not a single record (can happen if last line without
	    # trailing newline and reading backwards!)
	    $recordOK = 0;
	}

	if ($recordOK) {
	    $jid = $ARRAY[3]; # finished job, return LRMS ID!
	}
	else {
	    print "Ignoring entry from LSF log file (incomplete): $_[0]\n";
	}
    }
    else {
	print "Warning: LSF log file seems to have a wrong format! Expected: \"JOB_FINISH\" ...; having: $_[0]\n";
    }

    return $jid; # line to ignore?
}


# get event time for LRMS log entry: returns 0 if not found
sub getLrmsEventTime {
    if ($_[0] eq "pbs") {
	return &getLrmsEventTime_pbs($_[1], $_[2]);
    }
    elsif ($_[0] eq "lsf") {
	return &getLrmsEventTime_lsf($_[1], $_[2]);
    }
    return 0;
}

sub getLrmsEventTime_pbs {
    # Format in PBS log: 03/10/2006 00:03:33;E; ...

    my $eventTimestamp = 0;

    my @ARRAY = split(";" , $_[0] );
    if (scalar(@ARRAY) > 0) {
	my $sec = 0; my $min = 0; my $hour = 0;
	my $mday = 0; my $mon = 0; my $year = 0;

	if ($ARRAY[0] =~
	    /^(\d{2})\/(\d{2})\/(\d{4})\s(\d{2}):(\d{2}):(\d{2})$/) {

	    $mon = int($1)-1; # has to be 0 to 11 -> -1 !
	    $mday = int($2);
	    $year = int($3);
	    $hour = int($4);
	    $min = int($5);
	    $sec = int($6);

	    $eventTimestamp = timelocal($sec,$min,$hour,    # ss:mm:hh
					$mday,$mon,$year);  # dd-mm-yy
	    $_[1] = $ARRAY[0];
	}
    }

    return $eventTimestamp;
}

sub getLrmsEventTime_lsf {
    # Format in LSF log: "JOB_FINISH" "6.0" 1140194675 ...

    my $eventTimestamp = 0;

    my @ARRAY = split(" " , $_[0] );
    if ( (scalar(@ARRAY) > 2) && ($ARRAY[2] =~ /^(\d*)$/) ) {
	$eventTimestamp = int($1);

	$_[1] = $ARRAY[2];
    }

    return $eventTimestamp;
}


# get the LRMS creation time for the job: returns 0 if not found:
sub getLrmsJobCTime {
    if ($_[0] eq "pbs") {
	return &getLrmsJobCTime_pbs($_[1]);
    }
    elsif ($_[0] eq "lsf") {
	return &getLrmsJobCTime_lsf($_[1]);
    }
    return 0;
}

sub getLrmsJobCTime_pbs {
    my $ctime = 0;

    if ($_[0] =~ /\sctime=(\d*)\s/) {
	$ctime = int($1);
    }

    return $ctime;
}

sub getLrmsJobCTime_lsf {
    my $ctime = 0;

    # in lsb.acct the creation time is the submitTime, the 8th field
    my @ARRAY = split(" " , $_[0] );
    if ( (scalar(@ARRAY) > 7) && ($ARRAY[7] =~ /^(\d*)$/) ) {
	$ctime = int($1);
    }

    return $ctime;
}


## ----- print error and quit ----- ##

sub error {
    if (scalar(@_) > 0) {
	print "$_[0]";
    }

    # but don't leave the lock file behind ...
    print "Exiting...\n";
    if ( &delLock($collectorLockFileName) != 0 ) {	
	print "Error removing lock file.\n";
    }
    exit(1);
}


sub errorNoRemoveLock {
    if (scalar(@_) > 0) {
	print "$_[0]";
    }
    exit(1);
}


sub date2Timestamp {
    my $tstamp = 0;
    if ($_[0] =~ /^(\d{4})[-\/](\d{2})[-\/](\d{2})$/) {
	# get timestamp for this UTC time!
	$tstamp = timegm(0,0,0,    # ss:mm:hh
			 int($3),int($2)-1,int($1)); # dd-mm-yy
	                 # month should be from 0 to 11 => -1 !
    }
    return $tstamp;
}
